import os

import pkg_resources
from jinja2 import (ChoiceLoader, Environment, FileSystemLoader,
                    select_autoescape)
from markdown_it import MarkdownIt  # type: ignore
from markdown_it.renderer import RendererHTML  # type: ignore
from pydantic import BaseConfig  # pylint: disable=no-name-in-module
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import (BaseSettings, Extra,  # pylint: disable=no-name-in-module
                      Field)
from pygments import highlight  # type: ignore
from pygments.formatters.html import HtmlFormatter  # type: ignore
from pygments.lexers import get_lexer_by_name  # type: ignore
from pygments.styles import get_style_by_name  # type: ignore


def highlight_code(code: str, lang: str, _) -> str:
    """
    A function to highlight code using pygments.

    Arguments:
    code -- The code to be highlighted.
    lang -- The language of the code.
    """
    lexer = get_lexer_by_name(lang, stripall=True)
    formatter = HtmlFormatter(  # type: ignore
        style=get_style_by_name("github-dark"),  # type: ignore
    )  # type: ignore
    return highlight(code, lexer, formatter)  # type: ignore


def render_markdown(text: str) -> str:
    """
    A function to render markdown using markdown-it.

    Arguments:
    text -- The markdown text to be rendered.
    """
    md = MarkdownIt(renderer_cls=RendererHTML)
    md.options.html = True
    md.options.linkify = True
    md.options.typographer = True
    md.options.breaks = True
    md.options.highlight = highlight_code
    return md.render(text)


class SSRSettings(BaseSettings):
    """Default settings for Static Site Generator"""

    class Config(BaseConfig):
        env_file = ".env"
        env_file_encoding = "utf-8"
        extra = Extra.allow

    STATIC_DIR: str = Field(
        default_factory=lambda: os.path.join(os.getcwd(), "static"), env="STATIC_DIR"
    )
    TEMPLATE_DIR: str = Field(
        default_factory=lambda: os.path.join(os.getcwd(), "templates"),
        env="TEMPLATE_DIR",
    )
    MARKDOWN_DIR: str = Field(
        default_factory=lambda: os.path.join(os.getcwd(), "markdown"),
        env="MARKDOWN_DIR",
    )


class SSRRenderer(BaseModel):
    """
    A class for Rendering Static Site Generator templates.

    Attributes:
            title (str): The title for the site. Default is "AioFaunaSSG".

    .. note::
            This class uses Pydantic for data validation.

    """

    title: str = Field(default="AioFaunaSSG")

    class Config(BaseConfig):
        extra = Extra.allow
        arbitrary_types_allowed = True


    @property
    def loader(self) -> ChoiceLoader:
        """
        Returns a ChoiceLoader for the Jinja2 Environment.

        Returns:
                ChoiceLoader: A loader that can load templates from multiple sources.
        """
        templates_loader = FileSystemLoader(SSRSettings().TEMPLATE_DIR)
        layouts_loader = FileSystemLoader(
            pkg_resources.resource_filename("aiofauna", "ssg/layouts")
        )
        markdown_loader = FileSystemLoader(SSRSettings().MARKDOWN_DIR)

        return ChoiceLoader([templates_loader, layouts_loader, markdown_loader])

    @property
    def settings(self) -> SSRSettings:
        """
        Returns the settings for the SSG.

        Returns:
                SSGSettings: The settings for the SSG.
        """
        return SSRSettings()

    @property
    def jinja_env(self) -> Environment:
        """
        Returns a Jinja2 Environment.

        Returns:
                Environment: The Jinja2 Environment for template rendering.
        """
        return Environment(
            loader=self.loader,
            autoescape=select_autoescape(["html", "xml", "md", "markdown"]),
            enable_async=True,
        )

    async def render_html(self, template: str) -> str:
        """
        Renders a template using Jinja2.

        Args:
                template (str): The name of the template file.

        Returns:
                str: The rendered HTML as a string.
        """
        return await self.jinja_env.get_template(template).render_async(**self.dict())

    async def render_markdown(
        self, template: str, layout: str = "md_layout.html"
    ) -> str:
        """
        Renders a markdown file using a layout.

        Args:
                template (str): The name of the markdown file.
                layout (str, optional): The layout to use. Defaults to "md_layout.html".

        Returns:
                str: The rendered HTML as a string.
        """
        content = await self.render_html(template)
        return await self.jinja_env.get_template(layout).render_async(
            title=self.title, content=render_markdown(content)
        )


from __future__ import annotations

from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from typing import Any, List, Type, TypeVar
from uuid import UUID

from pydantic import BaseConfig  # pylint: disable=no-name-in-module
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import Extra  # pylint: disable=no-name-in-module

from ..faunadb.objects import FaunaTime, Ref


class Document(BaseModel):
    """Document
    Base document for all FaunaDB documents
    """

    class Config(BaseConfig):
        """Configuration defaults for all documents"""

        extra = Extra.allow
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: lambda v: v.astimezone().isoformat(),
            date: lambda v: v.isoformat(),
            time: lambda v: v.isoformat(),
            timedelta: lambda v: v.total_seconds(),
            UUID: lambda v: str(v),
            Decimal: lambda v: float(v),
            Ref: lambda v: v.id(),
            FaunaTime: lambda v: v.value,
            Enum: lambda v: v.value,
        }

    class Metadata:
        subclasses: List[Type[Document]] = []

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        if cls.__doc__ is None:
            cls.__doc__ = f"```json\n{cls.schema_json(indent=2)}\n```"
        cls.Metadata.subclasses.append(cls)


D = TypeVar("D", bound=Document)



from __future__ import annotations

import json
import os
import typing
from typing import Any

from aiohttp.client_exceptions import ClientConnectionError, ClientConnectorSSLError
from aiohttp.web_exceptions import HTTPException
from dotenv import load_dotenv
from pydantic import Field

from ..data.json import to_json
from ..faunadb.errors import FaunaException
from ..faunadb.objects import Expr
from ..utils import setup_logging
from .api_client import Client

load_dotenv()


class MissingEnvironmentVariable(Exception):
    """Exception for missing environment variable."""

    ...


if "FAUNA_SECRET" not in os.environ:
    raise MissingEnvironmentVariable(
        "The FAUNA_SECRET environment variable is not set."
    )


FAUNA_SECRET = os.environ["FAUNA_SECRET"]
HEADERS = {
    "Authorization": f"Bearer {FAUNA_SECRET}",
    "Content-type": "application/json",
    "Accept": "application/json",
    "User-Agent": "aiofauna-framework",
}

T = typing.TypeVar("T")

logger = setup_logging(__name__)

FAUNA_EXCEPTIONS = (
    HTTPException,
    FaunaException,
    ValueError,
    KeyError,
    TypeError,
    Exception,
    UnicodeError,
    json.JSONDecodeError,
    RuntimeError,
    ClientConnectionError,
    ClientConnectorSSLError,
)


class APIException(HTTPException):
    """
    API Exception
    -------------
    Args:
        message (str): Error message
        status_code (int, optional): HTTP status code. Defaults to 500.

    Raises:
        HTTPException: HTTP Exception
    """

    def __init__(self, message: str, status_code: int = 500) -> None:
        self.message = message
        self.status_code = status_code
        super().__init__(
            text=json.dumps({"message": message, "status": status_code}),
            content_type="application/json",
        )


class FaunaClient(Client):
    """
    FaunaDB Client
    --------------

    The main client for interacting with FaunaDB. This client is used to query the database and perform transactions on top of `FQL <https://docs.fauna.com/fauna/current/api/fql/>` language through a friendly pythonic interface.

    Args:
        base_url (str, optional): Base URL for the client. Defaults to "https://db.fauna.com".

    Attributes:
        base_url (str): Base URL for the client.
        headers (dict[str, str]): Headers for the client.

    """

    base_url: str = Field(default="https://db.fauna.com")
    headers: dict[str, str] = Field(default_factory=lambda: HEADERS)

    async def query(self, expr: Expr) -> Any:
        """
        Query FaunaDB
        -----------------
        Args:
            expr (Expr): FaunaDB expression

            A FaunaDB expression is a JSON object that represents a FaunaDB query. For more information on FaunaDB expressions, see the `FQL <https://docs.fauna.com/fauna/current/api/fql/>`_ documentation.

            Basically an `Expr` is a `Ref`, `SetRef` or `FaunaTime` object, wrapping a JSON object that contains a FaunaDB query.

        Returns:

            The result of the query.

        Raises:

            APIException: API Exception

        Example:

            .. code-block:: python

                from aiofauna import FaunaClient
                from aiofauna.data import query

                client = FaunaClient()

                await client.query(q.create_database({"name": "my_database"}))
        """

        response = await self().post("/", data=to_json(expr))
        try:
            data = await response.json()
            if data.get("resource") is not None:
                return data["resource"]
            if data.get("error") is not None:
                return data["error"]
            return data
        except FAUNA_EXCEPTIONS as exc:  # pylint:disable=all
            logger.error(exc.__class__.__name__)  # type: ignore
            logger.error(exc)
            raise APIException(str(exc)) from exc



"""REST API Module with automatic OpenAPI generation."""
import os
from functools import wraps
from inspect import signature
from typing import (Any, Awaitable, Callable, Coroutine, List, Literal,
                    Optional, TypeVar, cast)

from aiohttp.typedefs import Handler
from aiohttp.web import (Application, HTTPFound, Request, Response,
                         StreamResponse)
from aiohttp.web_middlewares import middleware
from aiohttp.web_ws import WebSocketResponse
from aiohttp_sse import EventSourceResponse, sse_response
from typing_extensions import ParamSpec

from ..data.document import Document
from ..data.json import jsonable_encoder
from ..data.odm import FaunaModel
from ..helpers import do_response
from ..utils import setup_logging
from .docs import extract, html_string, load, transform
from .router import APIRouter

P = ParamSpec("P")
_T = TypeVar("_T")

Middleware = Callable[[Request, Handler], Awaitable[StreamResponse]]


class APIServer(Application):
    """
    AioFauna APIServer class for handling API operations.

    :param title: The title of the API.
    :type title: str
    :param servers: List of server URLs.
    :type servers: list[str]
    :param description: API description.
    :type description: str
    :param version: API version.
    :type version: str
    :param openapi_url: URL for OpenAPI documentation.
    :type openapi_url: str
    """

    def __init__(
        self,
        *args: Any,
        title: str = "AioFauna",
        servers: Optional[List[str]] = None,
        version: str = "0.0.1",
        openapi_url: str = "/openapi.json",
        **kwargs: Any,
    ):
        super().__init__(*args, logger=setup_logging(self.__class__.__name__), **kwargs)
        schemas = FaunaModel.Metadata.subclasses + Document.Metadata.subclasses
        self.openapi = {
            "openapi": "3.1.0",
            "info": {"title": title, "version": version},
            "paths": {},
            "servers": [{"url": url} for url in servers or []],
            "components": {
                "schemas": {schema.__name__: schema.schema() for schema in schemas}
            },
        }
        self._route_open_api_params = {}
        self.openapi_url: str = openapi_url

        @self.get("/openapi.json")
        async def openapi():
            response = jsonable_encoder(self.openapi)
            return response

        @self.get("/docs")
        async def docs():
            return Response(
                text=html_string(self.openapi_url), content_type="text/html"
            )

    def document(
        self, path: str, method: str
    ) -> Callable[
        [Callable[P, Coroutine[Any, Any, _T]]], Callable[P, Coroutine[Any, Any, _T]]
    ]:
        """

        Decorator to generate OpenAPI documentation for an endpoint handler.

        :param path: The URL path of the endpoint.
        :type path: str
        :param method: The HTTP method of the endpoint.
        :type method: str
        :return: Decorated function for handling the endpoint.
        :rtype: Callable

        """

        def decorator(
            func: Callable[P, Coroutine[Any, Any, _T]]
        ) -> Callable[P, Coroutine[Any, Any, _T]]:
            sig = signature(func)
            params = sig.parameters
            open_api_params = extract(params.copy(), path)
            self._route_open_api_params[(path, method)] = open_api_params
            transform(self.openapi, path, method, func, open_api_params)

            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> StreamResponse:
                request = cast(Request, args[0])
                new_args = args[1:]
                args_to_apply = await load(request, params.copy())
                definitive_args = {}
                for name, param in params.items():
                    if name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                response = await func(*new_args, **definitive_args, **kwargs)  # type: ignore
                return do_response(response)

            wrapper._handler = func
            wrapper.__name__ = func.__name__
            return cast(Callable[P, Coroutine[Any, Any, _T]], wrapper)

        return decorator

    def get(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """
        Decorator for handling GET requests.

        :param path: The API path for the GET request.
        :type path: str
        :return: Decorated function for handling the GET request.
        :rtype: Callable
        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_get(path, self.document(path, "GET")(func), **kwargs)
            return func

        return decorator

    def post(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """
        Decorator for handling POST requests.

        :param path: The API path for the POST request.
        :type path: str
        :return: Decorated function for handling the POST request.
        :rtype: Callable
        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_post(path, self.document(path, "POST")(func), **kwargs)
            return func

        return decorator

    def put(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """
        Decorator for handling PUT requests.

        :param path: The API path for the PUT request.
        :type path: str
        :return: Decorated function for handling the PUT request.
        :rtype: Callable

        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_put(path, self.document(path, "PUT")(func), **kwargs)
            return func

        return decorator

    def delete(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """
        Decorator for handling DELETE requests.

        :param path: The API path for the DELETE request.
        :type path: str
        :return: Decorated function for handling the DELETE request.
        :rtype: Callable

        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_delete(path, self.document(path, "DELETE")(func), **kwargs)
            return func

        return decorator

    def patch(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """

        Decorator for handling PATCH requests.

        :param path: The API path for the PATCH request.
        :type path: str
        :return: Decorated function for handling the PATCH request.
        :rtype: Callable

        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_patch(path, self.document(path, "PATCH")(func), **kwargs)
            return func

        return decorator

    def head(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """

        Decorator for handling HEAD requests.

        :param path: The API path for the HEAD request.
        :type path: str
        :return: Decorated function for handling the HEAD request.
        :rtype: Callable

        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_head(path, self.document(path, "HEAD")(func), **kwargs)
            return func

        return decorator

    def options(
        self, path: str, **kwargs: Any
    ) -> Callable[[Callable[P, Awaitable[_T]]], Callable[P, Awaitable[_T]]]:
        """

        Decorator for handling OPTIONS requests.

        :param path: The API path for the OPTIONS request.
        :type path: str
        :return: Decorated function for handling the OPTIONS request.
        :rtype: Callable

        """

        def decorator(
            func: Callable[P, Awaitable[StreamResponse]]
        ) -> Callable[P, Awaitable[StreamResponse]]:
            self.router.add_options(
                path, self.document(path, "OPTIONS")(func), **kwargs
            )
            return func

        return decorator

    def on_event(
        self, event: Literal["startup", "shutdown"]
    ) -> Callable[
        [Callable[[Application], Awaitable[None]]],
        Callable[[Application], Awaitable[None]],
    ]:
        """

        Decorator for handling startup and shutdown lifecycle events.

        :param event: The event to handle (startup or shutdown).

        """

        def decorator(
            func: Callable[[Application], Awaitable[None]]
        ) -> Callable[[Application], Awaitable[None]]:
            if event not in ("startup", "shutdown"):
                raise ValueError("Event must be startup or shutdown")
            elif event == "startup":
                self.on_startup.append(func)
            else:
                self.on_shutdown.append(func)
            return func

        return decorator

    def sse(
        self, path: str
    ) -> Callable[
        [Callable[P, Awaitable[EventSourceResponse]]],
        Callable[P, Awaitable[EventSourceResponse]],
    ]:
        """

        Server-Sent Events decorator.

        :param path: The API path for the SSE request.
        :type path: str
        :return: Decorated function for handling the SSE request.

        This decorator will handle SSE connections and pass the response object to the decorated function, be sure to include a parameter with the type annotation `EventSourceResponse` in your function signature.
        You can use the `send` method of the response object to send events to the client.

        Example:

        .. code-block:: python

        import asyncio
        from aiofauna import FaunaClient, APIServer, EventSourceResponse, to_json,q

        app = APIServer()
        fauna = FaunaClient()

        @app.sse("/stream")
        async def stream(sse: EventSourceResponse):
                        while True:
                                        now = await fauna.query(q.now())
                                        await sse.send(to_json(now["@ts"]))
                                        await asyncio.sleep(1)

        """

        def decorator(
            func: Callable[P, Awaitable[EventSourceResponse]]
        ) -> Callable[P, Awaitable[EventSourceResponse]]:
            @wraps(func)
            async def wrapper(request: Request) -> EventSourceResponse:
                async with sse_response(request) as resp:
                    args_to_apply = await load(
                        request, signature(func).parameters.copy()
                    )
                    definitive_args = {}
                    for name, param in signature(func).parameters.items():
                        if param.annotation == EventSourceResponse:
                            definitive_args[name] = resp
                        elif name in args_to_apply:
                            definitive_args[name] = args_to_apply[name]
                            args_to_apply.pop(name)
                        elif param.default is not param.empty:
                            definitive_args[name] = param.default
                        else:
                            raise ValueError(
                                f"Missing parameter {name} for {func.__name__}"
                            )
                    await func(**definitive_args)
                    return resp

            self.router.add_get(path, self.document(path, "OPTIONS")(wrapper))
            return wrapper

        return decorator

    def websocket(
        self, path: str
    ) -> Callable[
        [Callable[P, Awaitable[WebSocketResponse]]],
        Callable[P, Awaitable[WebSocketResponse]],
    ]:
        """

        Websocket decorator.

        :param path: The API path for the websocket request.
        :type path: str
        :return: Decorated function for handling the websocket request.

        This decorator will handle websocket connections and pass the websocket object to the decorated function, be sure to include a parameter with the type annotation `WebSocketResponse` in your function signature and also setup proper error handling, since `PING` and `PONG` messages are handled automatically by the server and will result on a `MessageTypeError` if not handled.

        """

        def decorator(
            func: Callable[P, Awaitable[WebSocketResponse]]
        ) -> Callable[P, Awaitable[WebSocketResponse]]:
            @wraps(func)
            async def wrapper(request: Request):
                args_to_apply = await load(request, signature(func).parameters.copy())
                ws = WebSocketResponse()
                await ws.prepare(request)
                definitive_args = {}
                for name, param in signature(func).parameters.items():
                    if param.annotation == WebSocketResponse:
                        definitive_args[name] = ws
                    elif name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                        args_to_apply.pop(name)
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                await func(**definitive_args)
                return ws

            self.router.add_get(path, wrapper)
            return wrapper

        return decorator

    def static(self, path: str = "", dir: str = "static") -> "APIServer":
        """

        Static files decorator.

        This decorator will serve static files from the `static` directory.
        It will also serve the `index.html` file from the `static` directory when the root path is requested.
        Useful for serving a frontend single-page application.

        :return: The server instance.

        """

        @self.get(f"/{path}")
        async def index(request: Request):
            base_url = request.url.path
            return HTTPFound(location=f"{base_url}/{path}/index.html")

        try:
            os.makedirs(dir, exist_ok=True)
        except OSError:
            pass
        self.router.add_static(f"/{path}", dir)

        return self

    def middleware(self, func: Middleware) -> Middleware:
        @wraps(func)
        @middleware
        async def wrapper(request: Request, handler: Handler) -> Response:
            response = await func(request, handler)
            if isinstance(response, Response):
                return response
            return do_response(response)

        self.middlewares.append(wrapper)
        return wrapper

    def use(self, router: APIRouter, prefix: str = "") -> "APIServer":
        """Chainable method to add a router to the server"""
        for route in router.routes:
            path = prefix + route.path
            self.router.add_route(
                route.method,
                prefix + router.prefix + path,
                route.handler,
                **route.kwargs,
            )

        for path, details in router.openapi["paths"].items():
            path = prefix + path
            self.openapi["paths"][path] = details

        self.openapi["components"]["schemas"].update(
            router.openapi["components"]["schemas"]
        )
        return self
